# HT-PAMDA normcount2rate (script 3)
#calculate rates (k) of PAM depletion, modeled as exponential decay y(t)=e^-(k*t)

#----------------------------------------------------------------------------------------------#

#input files and parameters:

import pandas as pd
import numpy as np
import math
from scipy.optimize import curve_fit

pam_length = 4

timepoint_sec = [0, 60, 480, 1920] # timepoints used (seconds)

read_sum_minimum = 10 # minimum read count to determine rate constant (if below limit: k = NaN)
tps_sum_minimum = 2 # minimum non-zero timepoints to determine rate constant (if below limit: k = NaN)

#calculate rates and generate heatmaps using these timepoints:
use_timepoints = [[0,1,2,3]] 

#exponential decay
def func(x, a, b):
        return a*np.exp(-b*x)

#----------------------------------------------------------------------------------------------#

# Find rates:

for timepoints in use_timepoints:

    print 'finding rate constants k with timepoints ', timepoints

    df = pd.read_csv('pam_depletion_output_norm_counts_%sntPAM.csv' % pam_length) #NAME OF INPUT FILE

    ks = []
    timepoint_indices = []
    for i in timepoints:
        timepoint_indices.append([i, timepoint_sec[i]])

    previous_row = 'n/a'

    #loop through one row at a time and nonlinear fit rates
    for index, row in df.iterrows():

        current_row = row['Sample']
        if current_row != previous_row:
            print current_row
        
        tps = [x[1] for x in timepoint_indices]
        obs_raw = [0.00000000001] + [row['Raw_Counts_' + str(x[0])] for x in timepoint_indices if str(x[0]) != '0']
        obs_norm = [row['Norm_Counts_' + str(x[0])] for x in timepoint_indices]

        zero_indices = [i for i, e in enumerate(obs_raw) if e == 0]
        for zero_index in sorted(zero_indices)[::-1]:
            del tps[zero_index]
            del obs_norm[zero_index]
            del obs_raw[zero_index]

        if sum(obs_raw) >= read_sum_minimum and len(obs_norm) >= tps_sum_minimum:

            # map initial conditions to fit error
            min_search = []
            
            # loop across different initial conditions for k
            for j in [0.0001, 0.001, 0.01]:
                p0 = [1.0, j]
                # nonlinear curve fit
                popt, pcov = curve_fit(func, tps, obs_norm, p0 = p0)
                # prediction from outputted paramters
                pred = [func(x, popt[0], popt[1]) for x in tps]
                # error
                error = sum([(x[0]-x[1])**2 for x in zip(pred, obs_norm)])
                # append into dictionary mapping initial conditions to fit error
                min_search.append([error, list(popt)])
                
            ks.append(sorted(min_search)[0][1][1])
 
        else:
            ks.append('NaN')
        previous_row = current_row

    print "appending rate constants"

    df['Rate_Constant_k'] = [x if (x > 0 or x == 'NaN') else min([x if x>0 else 100 for x in ks]) for x in ks]

    print "output to CSV"

    df.to_csv('pam_depletion_output_rates_%sntPAM_%s.csv' % (pam_length, timepoints) , index = False)